@@ -3,17 +3,27 @@
 '''
 A library to access Tryton's models like a client.
 '''
-__version__ = "2.2.2"
+__version__ = "2.8.0"
 __all__ = ['Model', 'Wizard']
-from types import NoneType
+import sys
+try:
+    import cdecimal
+    # Use cdecimal globally
+    if 'decimal' not in sys.modules:
+        sys.modules['decimal'] = cdecimal
+except ImportError:
+    import decimal
+    sys.modules['cdecimal'] = decimal
 import threading
 import datetime
 from decimal import Decimal
+from types import NoneType
+
 import proteus.config
-from proteus.pyson import PYSONDecoder
 
 _MODELS = threading.local()
 
+
 class _EvalEnvironment(dict):
     'Dictionary for evaluation'
     def __init__(self, parent):
@@ -58,12 +68,12 @@ class FieldDescriptor(object):
         self.__doc__ = definition['string']
 
     def __get__(self, instance, owner):
-        if instance.id > 0 and self.name not in instance._values:
+        if instance.id > 0:
             instance._read(self.name)
         return instance._values.get(self.name, self.default)
 
     def __set__(self, instance, value):
-        if instance.id > 0 and self.name not in instance._values:
+        if instance.id > 0:
             instance._read(self.name)
         previous = getattr(instance, self.name)
         instance._values[self.name] = value
@@ -90,35 +100,42 @@ class CharDescriptor(FieldDescriptor):
         assert isinstance(value, basestring) or value in (None, False)
         super(CharDescriptor, self).__set__(instance, value or '')
 
+    def __get__(self, instance, owner):
+        value = super(CharDescriptor, self).__get__(instance, owner)
+        if value is False:
+            value = None
+        return value
+
 
 class BinaryDescriptor(FieldDescriptor):
     default = None
 
     def __set__(self, instance, value):
-        assert isinstance(value, (basestring, buffer)) or value in (None, False)
+        assert (isinstance(value, (basestring, buffer))
+            or value in (None, False))
         super(BinaryDescriptor, self).__set__(instance, value or '')
 
 
 class IntegerDescriptor(FieldDescriptor):
-    default = 0
 
     def __set__(self, instance, value):
-        assert isinstance(value, (int, long))
+        assert isinstance(value, (int, long, NoneType))
         super(IntegerDescriptor, self).__set__(instance, value)
 
 
 class FloatDescriptor(FieldDescriptor):
-    default = 0.0
 
     def __set__(self, instance, value):
-        super(FloatDescriptor, self).__set__(instance, float(value))
+        assert isinstance(value, (int, long, float, Decimal, NoneType))
+        if value is not None:
+            value = float(value)
+        super(FloatDescriptor, self).__set__(instance, value)
 
 
 class NumericDescriptor(FieldDescriptor):
-    default = Decimal('0.0')
 
     def __set__(self, instance, value):
-        assert isinstance(value, Decimal)
+        assert isinstance(value, (NoneType, Decimal))
         # TODO add digits validation
         super(NumericDescriptor, self).__set__(instance, value)
 
@@ -163,6 +180,18 @@ class DateTimeDescriptor(FieldDescriptor):
         super(DateTimeDescriptor, self).__set__(instance, value)
 
 
+class TimeDescriptor(FieldDescriptor):
+    def __set__(self, instance, value):
+        assert isinstance(value, datetime.time) or value in (None, False)
+        super(TimeDescriptor, self).__set__(instance, value)
+
+
+class DictDescriptor(FieldDescriptor):
+    def __set__(self, instance, value):
+        assert isinstance(value, dict) or value in (None, False)
+        super(DictDescriptor, self).__set__(instance, value)
+
+
 class Many2OneDescriptor(FieldDescriptor):
     def __get__(self, instance, owner):
         relation = Model.get(self.definition['relation'], instance._config)
@@ -171,7 +200,8 @@ class Many2OneDescriptor(FieldDescriptor):
             value = relation(value)
         elif not value:
             value = None
-        instance._values[self.name] = value
+        if self.name in instance._values:
+            instance._values[self.name] = value
         return value
 
     def __set__(self, instance, value):
@@ -190,12 +220,17 @@ class One2ManyDescriptor(FieldDescriptor):
     default = []
 
     def __get__(self, instance, owner):
+        from .pyson import PYSONDecoder
         relation = Model.get(self.definition['relation'], instance._config)
         value = super(One2ManyDescriptor, self).__get__(instance, owner)
         if not isinstance(value, ModelList):
-            value = ModelList((relation(id)
-                for id in value or []), instance, self.name,
-                self.definition.get('relation_field'))
+            ctx = instance._context.copy() if instance._context else {}
+            if self.definition.get('context'):
+                decoder = PYSONDecoder(_EvalEnvironment(instance))
+                ctx.update(decoder.decode(self.definition.get('context')))
+            with instance._config.set_context(ctx):
+                value = ModelList(self.definition, (relation(id)
+                        for id in value or []), instance, self.name)
             instance._values[self.name] = value
         return value
 
@@ -222,13 +257,13 @@ class ReferenceValueDescriptor(ValueDescriptor):
         value = super(ReferenceValueDescriptor, self).__get__(instance, owner)
         if isinstance(value, Model):
             value = '%s,%s' % (value.__class__.__name__, value.id)
-        return value or False
+        return value or None
 
 
 class Many2OneValueDescriptor(ValueDescriptor):
     def __get__(self, instance, owner):
         value = super(Many2OneValueDescriptor, self).__get__(instance, owner)
-        return value and value.id or False
+        return value and value.id or None
 
 
 class One2OneValueDescriptor(Many2OneValueDescriptor):
@@ -239,14 +274,17 @@ class One2ManyValueDescriptor(ValueDescriptor):
     def __get__(self, instance, owner):
         value = [('add', [])]
         value_list = getattr(instance, self.name)
+        to_create = []
         for record in value_list:
             if record.id > 0:
                 if record._changed:
-                    value.append(('write', record.id, record._get_values(
+                    value.append(('write', [record.id], record._get_values(
                         fields=record._changed)))
                 value[0][1].append(record.id)
             else:
-                value.append(('create', record._get_values()))
+                to_create.append(record._get_values())
+        if to_create:
+            value.append(('create', to_create))
         if value_list.record_removed:
             value.append(('unlink', [x.id for x in value_list.record_removed]))
         if value_list.record_deleted:
@@ -273,7 +311,7 @@ class ReferenceEvalDescriptor(EvalDescriptor):
         value = super(ReferenceEvalDescriptor, self).__get__(instance, owner)
         if isinstance(value, Model):
             value = '%s,%s' % (value.__class__.__name__, value.id)
-        return value or False
+        return value or None
 
 
 class Many2OneEvalDescriptor(EvalDescriptor):
@@ -281,7 +319,7 @@ class Many2OneEvalDescriptor(EvalDescriptor):
         value = super(Many2OneEvalDescriptor, self).__get__(instance, owner)
         if value:
             return value.id
-        return False
+        return None
 
 
 class One2OneEvalDescriptor(Many2OneEvalDescriptor):
@@ -304,7 +342,7 @@ class MetaModelFactory(object):
         'text': CharDescriptor,
         'sha': CharDescriptor,
         'binary': BinaryDescriptor,
-        'selection': CharDescriptor, # TODO implement its own descriptor
+        'selection': CharDescriptor,  # TODO implement its own descriptor
         'integer': IntegerDescriptor,
         'biginteger': IntegerDescriptor,
         'float': FloatDescriptor,
@@ -313,6 +351,8 @@ class MetaModelFactory(object):
         'reference': ReferenceDescriptor,
         'date': DateDescriptor,
         'datetime': DateTimeDescriptor,
+        'time': TimeDescriptor,
+        'dict': DictDescriptor,
         'many2one': Many2OneDescriptor,
         'one2many': One2ManyDescriptor,
         'many2many': Many2ManyDescriptor,
@@ -342,6 +382,7 @@ class MetaModelFactory(object):
         models_key = 'c%s' % id(self.config)
         if not hasattr(_MODELS, models_key):
             setattr(_MODELS, models_key, {})
+
         class MetaModel(type):
             'Meta class for Model'
             def __new__(mcs, name, bases, dict):
@@ -378,22 +419,26 @@ class MetaModelFactory(object):
 class ModelList(list):
     'List for Model'
 
-    def __init__(self, sequence=None, parent=None, parent_field_name='',
-            parent_name=''):
+    def __init__(self, definition, sequence=None, parent=None,
+            parent_field_name=''):
+        self.model_name = definition['relation']
         if sequence is None:
             sequence = []
         self.parent = parent
         if parent:
             assert parent_field_name
         self.parent_field_name = parent_field_name
-        self.parent_name = parent_name
+        self.parent_name = definition.get('relation_field', '')
+        self.domain = definition.get('domain', [])
+        self.context = definition.get('context')
+        self.add_remove = definition.get('add_remove')
         self.record_removed = set()
         self.record_deleted = set()
         result = super(ModelList, self).__init__(sequence)
         for record in self:
             record._parent = parent
             record._parent_field_name = parent_field_name
-            record._parent_name = parent_name
+            record._parent_name = self.parent_name
         return result
     __init__.__doc__ = list.__init__.__doc__
 
@@ -403,6 +448,13 @@ class ModelList(list):
             self.parent._changed.add(self.parent_field_name)
             self.parent._on_change(self.parent_field_name)
 
+    def _get_context(self):
+        from .pyson import PYSONDecoder
+        decoder = PYSONDecoder(_EvalEnvironment(self.parent))
+        ctx = self.parent._context.copy() if self.parent._context else {}
+        ctx.update(decoder.decode(self.context) if self.context else {})
+        return ctx
+
     def append(self, record):
         assert isinstance(record, Model)
         if self.parent:
@@ -477,6 +529,28 @@ class ModelList(list):
         raise NotImplementedError
     sort.__doc__ = list.sort.__doc__
 
+    def new(self):
+        'Adds a new record to the ModelList and returns it'
+        Relation = Model.get(self.model_name, self.parent._config)
+        with Relation._config.set_context(self._get_context()):
+            new_record = Relation()
+        self.append(new_record)
+        return new_record
+
+    def find(self, condition=None, offset=0, limit=None, order=None):
+        'Returns records matching condition taking into account list domain'
+        from .pyson import PYSONDecoder
+        decoder = PYSONDecoder(_EvalEnvironment(self.parent))
+        Relation = Model.get(self.model_name, self.parent._config)
+        if condition is None:
+            condition = []
+        field_domain = decoder.decode(self.domain)
+        add_remove_domain = (decoder.decode(self.add_remove)
+            if self.add_remove else [])
+        new_domain = [field_domain, add_remove_domain, condition]
+        with Relation._config.set_context(self._get_context()):
+            return Relation.find(new_domain, offset, limit, order)
+
 
 class Model(object):
     'Model class for Tryton records'
@@ -493,11 +567,12 @@ class Model(object):
         self.__id = id or Model.__counter
         if self.__id < 0:
             Model.__counter -= 1
-        self._values = {} # store the values of fields
-        self._changed = set() # store the changed fields
-        self._parent = None # store the parent record
-        self._parent_field_name = '' # store the field name in parent record
-        self._parent_name = '' # store the field name to parent record
+        self._values = {}  # store the values of fields
+        self._changed = set()  # store the changed fields
+        self._parent = None  # store the parent record
+        self._parent_field_name = ''  # store the field name in parent record
+        self._parent_name = ''  # store the field name to parent record
+        self._context = self._config.context  # store the context
         if self.id < 0 and _default:
             self._default_get()
 
@@ -521,6 +596,9 @@ class Model(object):
     @classmethod
     def get(cls, name, config=None):
         'Get a class for the named Model'
+        if isinstance(name, unicode):
+            name = name.encode('utf-8')
+
         class Spam(Model):
             __metaclass__ = MetaModelFactory(name, config=config)()
         return Spam
@@ -580,14 +658,14 @@ class Model(object):
         'Save the record'
         context = self._config.context
         if self.id < 0:
-            values = self._get_values(fields=self._changed)
-            self.__id = self._proxy.create(values, context)
+            values = self._get_values()
+            self.__id, = self._proxy.create([values], context)
         else:
             if not self._changed:
                 return
             values = self._get_values(fields=self._changed)
             context['_timestamp'] = self._get_timestamp()
-            self._proxy.write(self.id, values, context)
+            self._proxy.write([self.id], values, context)
         self.reload()
 
     def delete(self):
@@ -595,7 +673,7 @@ class Model(object):
         if self.id > 0:
             context = self._config.context
             context['_timestamp'] = self._get_timestamp()
-            return self._proxy.delete(self.id, context)
+            return self._proxy.delete([self.id], context)
         self.reload()
         return True
 
@@ -625,14 +703,17 @@ class Model(object):
 
     def _read(self, name):
         'Read field'
-        fields = []
-        if not self._values:
+        fields = [name]
+        if name in self._values:
+            return
+        loading = self._fields[name]['loading']
+        if loading == 'eager':
             fields = [x for x, y in self._fields.iteritems()
-                    if y['type'] not in ('one2many', 'many2many', 'binary')]
+                    if y['loading'] == 'eager']
+        if not self._fields:
             fields.append('_timestamp')
-        fields.append(name)
-        self._values.update(self._proxy.read(self.id, fields,
-            self._config.context))
+        self._values.update(self._proxy.read([self.id], fields,
+            self._config.context)[0])
         for field in fields:
             if (field in self._fields
                     and self._fields[field]['type'] == 'float'
@@ -648,6 +729,8 @@ class Model(object):
 
     def _default_set(self, values):
         for field, value in values.iteritems():
+            if '.' in field:
+                continue
             definition = self._fields[field]
             if definition['type'] in ('one2many', 'many2many'):
                 if value and len(value) and isinstance(value[0], (int, long)):
@@ -659,8 +742,8 @@ class Model(object):
                     record = relation()
                     record._default_set(vals)
                     records.append(record)
-                self._values[field] = ModelList(records, self, field,
-                        definition.get('relation_field', ''))
+                self._values[field] = ModelList(definition, records, self,
+                    field)
             else:
                 self._values[field] = value
 
@@ -670,16 +753,20 @@ class Model(object):
         values['id'] = self.id
         return values
 
+    def _get_on_change_value(self):
+        values = self._get_eval()
+        for field, definition in self._fields.iteritems():
+            if definition['type'] == 'one2many':
+                values[field] = [x._get_on_change_value()
+                    for x in getattr(self, field)]
+        return values
+
     def _on_change_args(self, args):
         res = {}
-        values = self._get_eval()
-        del values['id']
-        for field, definition in self._fields.iteritems():
-            if definition['type'] in ('one2many', 'many2many'):
-                values[field] = [x._get_eval() for x in getattr(self, field)]
+        values = self._get_on_change_value()
         if self._parent:
             values['_parent_%s' % self._parent_name] = \
-                    _EvalEnvironment(self._parent)
+                _EvalEnvironment(self._parent)
         for arg in args:
             scope = values
             for i in arg.split('.'):
@@ -730,6 +817,8 @@ class Model(object):
 
     def _on_change(self, name):
         'Call on_change for field'
+        # Import locally to not break installation
+        from proteus.pyson import PYSONDecoder
         definition = self._fields[name]
         if definition.get('on_change'):
             if isinstance(definition['on_change'], basestring):
@@ -748,6 +837,9 @@ class Model(object):
                 self._on_change_set(field, value)
             for field, value in later.iteritems():
                 self._on_change_set(field, value)
+        values = {}
+        to_change = set()
+        later = set()
         for field, definition in self._fields.iteritems():
             if not definition.get('on_change_with'):
                 continue
@@ -755,11 +847,23 @@ class Model(object):
                 continue
             if field == name:
                 continue
-            args = self._on_change_args(definition['on_change_with'])
+            if to_change & set(definition['on_change_with']):
+                later.add(field)
+            to_change.add(field)
+            values.update(self._on_change_args(definition['on_change_with']))
+        if to_change:
             context = self._config.context
-            res = getattr(self._proxy, 'on_change_with_%s' % field)(args,
+            result = getattr(self._proxy, 'on_change_with')(values,
+                list(to_change), context)
+            for field, value in result.items():
+                self._on_change_set(field, value)
+        for field in later:
+            definition = self._fields[field]
+            values = self._on_change_args(definition['on_change_with'])
+            context = self._config.context
+            result = getattr(self._proxy, 'on_change_with_%s' % field)(values,
                     context)
-            self._on_change_set(field, res)
+            self._on_change_set(field, result)
         if self._parent:
             self._parent._changed.add(self._parent_field_name)
             self._parent._on_change(self._parent_field_name)
@@ -773,62 +877,57 @@ class Wizard(object):
             assert len(set(type(x) for x in models)) == 1
         super(Wizard, self).__init__()
         self.name = name
-        self.state = None
-        self.states = ['init']
         self.form = None
+        self.form_state = None
         self._config = config or proteus.config.get_config()
         self._context = context or {}
         self._proxy = self._config.get_proxy(name, type='wizard')
-        self.id = self._proxy.create(self._config.context)
-        if models:
-            self.datas = {
-                    'model': models[0].__class__.__name__,
-                    'id': models[0].id,
-                    'ids': [model.id for model in models]
-                    }
-        else:
-            self.datas = {}
-        self.execute('init')
+        result = self._proxy.create(self._config.context)
+        self.session_id, self.start_state, self.end_state = result
+        self.states = [self.start_state]
+        self.models = models
+        self.execute(self.start_state)
 
     def execute(self, state):
         assert state in self.states
 
-        if 'form' not in self.datas:
-            self.datas['form'] = {}
-
-        if self.form:
-            self.datas['form'].update(self.form._get_values())
-
         self.state = state
-        while self.state != 'end':
+        while self.state != self.end_state:
             ctx = self._context.copy()
             ctx.update(self._config.context)
-            ctx['active_id'] = self.datas.get('id')
-            ctx['active_ids'] = self.datas.get('ids')
+            if self.models:
+                ctx['active_id'] = self.models[0].id
+                ctx['active_ids'] = [model.id for model in self.models]
+                ctx['active_model'] = self.models[0].__class__.__name__
+            else:
+                ctx['active_id'] = None
+                ctx['active_ids'] = None
+                ctx['active_model'] = None
 
-            res = self._proxy.execute(self.id, self.datas, self.state, ctx)
-            if not res:
-                break
+            if self.form:
+                # Filter only modified values
+                data = {self.form_state:
+                    dict((k, v) for k, v in
+                        self.form._get_on_change_value().iteritems()
+                        if k in self.form._values)}
+            else:
+                data = {}
 
-            if 'datas' in res:
-                self.datas['form'] = res['datas']
-            elif res['type'] == 'form':
-                self.datas['form'] = {}
+            result = self._proxy.execute(self.session_id, data, self.state,
+                ctx)
 
-            if res['type'] == 'form':
-                self.states = [x[0] for x in res['state']]
-                # XXX set context
-                self.form = Model.get(res['object'])()
-                self.form._default_set(self.datas['form'])
+            if 'view' in result:
+                view = result['view']
+                self.form = Model.get(view['fields_view']['model'])()
+                self.form._default_set(view['defaults'])
+                self.states = [b['state'] for b in view['buttons']]
+                self.form_state = view['state']
                 return
-            elif res['type'] == 'action':
-                # TODO run action
-                self.state = res['state']
-            elif res['type'] == 'print':
-                # TODO run print
-                self.state = res['state']
-            elif res['type'] == 'state':
-                self.state = res['state']
+            else:
+                self.state = self.end_state
 
-        if self.state == 'end':
-            self._proxy.delete(self.id, self._config.context)
+            if 'actions' in result:
+                pass  # TODO
+
+        if self.state == self.end_state:
+            self._proxy.delete(self.session_id, self._config.context)
